#include <assert.h>

#include "libguile/private-gc.h"
#include "libguile/gc.h"

#include "private-gc.h"

scm_t_freelist scm_master_freelist;
scm_t_freelist scm_master_freelist2;



#ifdef GUILE_DEBUG_FREELIST
/* Non-zero if freelist debugging is in effect.  Set this via
   `gc-set-debug-check-freelist!'. */
static int scm_debug_check_freelist = 0;
#endif

/* Debugging functions.  */

#if defined (GUILE_DEBUG) || defined (GUILE_DEBUG_FREELIST)


#ifdef GUILE_DEBUG_FREELIST
static void
map_free_list (scm_t_freelist *master, SCM freelist)
{
  abort();
  
  long last_seg = -1, count = 0;
  SCM f;

  for (f = freelist; !SCM_NULLP (f); f = SCM_FREE_CELL_CDR (f))
    {
      long int this_seg = scm_find_heap_segment_containing_object (f);

      if (this_seg == -1)
	{
	  fprintf (stderr,
		   "map_free_list: can't find segment containing cell %lux\n",
		   (unsigned long int) SCM_UNPACK (f));
	  abort ();
	}
      else if (this_seg != last_seg)
	{
	  if (last_seg != -1)
	    fprintf (stderr, "  %5ld %d-cells in segment %ld\n",
		     (long) count, master->span, (long) last_seg);
	  last_seg = this_seg;
	  count = 0;
	}
      count++;
    }
  if (last_seg != -1)
    fprintf (stderr, "  %5ld %d-cells in segment %ld\n",
	     (long) count, master->span, (long) last_seg);
}

SCM_DEFINE (scm_map_free_list, "map-free-list", 0, 0, 0,
            (),
	    "Print debugging information about the free-list.\n"
	    "@code{map-free-list} is only included in\n"
	    "@code{--enable-guile-debug} builds of Guile.")
#define FUNC_NAME s_scm_map_free_list
{
  size_t i;
  abort ();
  
  fprintf (stderr, "%ld segments total (%d:%ld",
	   (long) scm_heap_segment_table_size,
	   scm_heap_segment_table[0].span,
	   (long) (scm_heap_segment_table[0].bounds[1] - scm_heap_segment_table[0].bounds[0]));

  for (i = 1; i != scm_heap_segment_table_size; i++)
    fprintf (stderr, ", %d:%ld",
	     scm_heap_segment_table[i].span,
	     (long) (scm_heap_segment_table[i].bounds[1] - scm_heap_segment_table[i].bounds[0]));
  fprintf (stderr, ")\n");
  map_free_list (&scm_master_freelist, scm_freelist);
  map_free_list (&scm_master_freelist2, scm_freelist2);
  fflush (stderr);

  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


#endif /* defined (GUILE_DEBUG) || defined (GUILE_DEBUG_FREELIST) */

/*
  We used to have free list functions. Those are removed, since
  freelists don't apply with lazy sweeping.
 */

SCM_DEFINE (scm_gc_set_debug_check_freelist_x, "gc-set-debug-check-freelist!", 1, 0, 0,
            (SCM flag),
	    "If @var{flag} is @code{#t}, check the freelist for consistency\n"
	    "on each cell allocation.  This procedure only exists when the\n"
	    "@code{GUILE_DEBUG_FREELIST} compile-time flag was selected.")
#define FUNC_NAME s_scm_gc_set_debug_check_freelist_x
{
  /* [cmm] I did a double-take when I read this code the first time.
     well, FWIW. */
  SCM_VALIDATE_BOOL_COPY (1, flag, scm_debug_check_freelist);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

#endif /* GUILE_DEBUG_FREELIST */





/*
  This adjust FREELIST variables to decide wether or not to allocate
  more heap in the next GC run. It uses scm_gc_yield and scm_gc_yield1
 */

void
scm_adjust_min_yield (scm_t_freelist *freelist)
{
  /* min yield is adjusted upwards so that next predicted total yield
   * (allocated cells actually freed by GC) becomes
   * `min_yield_fraction' of total heap size.  Note, however, that
   * the absolute value of min_yield will correspond to `collected'
   * on one master (the one which currently is triggering GC).
   *
   * The reason why we look at total yield instead of cells collected
   * on one list is that we want to take other freelists into account.
   * On this freelist, we know that (local) yield = collected cells,
   * but that's probably not the case on the other lists.
   *
   * (We might consider computing a better prediction, for example
   *  by computing an average over multiple GC:s.)
   */
  if (freelist->min_yield_fraction)
    {
      /* Pick largest of last two yields. */
      long delta = ((SCM_HEAP_SIZE * freelist->min_yield_fraction / 100)
		   - (long) SCM_MAX (scm_gc_yield_1, scm_gc_yield));
#ifdef DEBUGINFO
      fprintf (stderr, " after GC = %lu, delta = %ld\n",
	       (long) scm_cells_allocated,
	       (long) delta);
#endif
      if (delta > 0)
	freelist->min_yield += delta;
    }
}


static void
scm_init_freelist (scm_t_freelist *freelist,
	       int span,
	       int min_yield)
{
  freelist->heap_segment_idx = -1;
  freelist->min_yield = 0;
  freelist->min_yield_fraction = min_yield;
  freelist->span = span;
  freelist->collected = 0;
  freelist->collected_1 = 0;
  freelist->heap_size = 0;
}


void
scm_gc_init_freelist (void)
{
  size_t init_heap_size_1
    = scm_getenv_int ("GUILE_INIT_SEGMENT_SIZE_1", scm_default_init_heap_size_1);

  size_t init_heap_size_2
    = scm_getenv_int ("GUILE_INIT_SEGMENT_SIZE_2", scm_default_init_heap_size_2);

  scm_freelist = SCM_EOL;
  scm_freelist2 = SCM_EOL;
  
  scm_init_freelist (&scm_master_freelist2, 2, 
		     scm_getenv_int ("GUILE_MIN_YIELD_2", scm_default_min_yield_2));
  scm_init_freelist (&scm_master_freelist, 1,
		     scm_getenv_int ("GUILE_MIN_YIELD_1", scm_default_min_yield_1));


  scm_max_segment_size = scm_getenv_int ("GUILE_MAX_SEGMENT_SIZE", scm_default_max_segment_size);
  
  scm_make_initial_segment (init_heap_size_1, &scm_master_freelist);
  scm_make_initial_segment (init_heap_size_2, &scm_master_freelist2);
}


void
scm_gc_sweep_freelist_reset (scm_t_freelist *freelist)
{
  freelist->collected_1 = freelist->collected;
  freelist->collected = 0;
  
  /*
    at the end we simply start with the lowest segment again.
   */
  freelist->heap_segment_idx = -1;
}

int
scm_gc_grow_heap_p (scm_t_freelist * freelist)
{
  return SCM_MAX (freelist->collected,freelist->collected_1)  < freelist->min_yield;
}
